__all__ = ["TR73U"]

import astropy.units as u

import ogameasure

from ... import config
from .weather_base import Weather


class TR73U(Weather):

    Manufacturer = "TandD"
    Model = "TR73U"

    def __init__(self) -> None:
        self.ondotori = ogameasure.TandD.tr_73u(config.antenna_tr73u_port)

    def get_temp(self) -> u.Quantity:
        data = self.ondotori.output_current_data()
        data_K = (data["temp"] * u.deg_C).to(u.K, equivalencies=u.temperature())
        return data_K

    def get_humid(self) -> float:
        data = self.ondotori.output_current_data() * 0.01
        return data["humid"]

    def get_press(self) -> u.Quantity:
        data = self.ondotori.output_current_data()
        return data["press"] * u.hPa

    def finalize(self) -> None:
        pass



from typing import Literal

from abc import ABC, abstractmethod


class PulseController(ABC):

    Manufacturer: str = ""
    Model: str

    @abstractmethod
    def set_step(self, step: int, axis: Literal["az", "el"]) -> None:
        ...

    @abstractmethod
    def set_speed(self, speed: float, axis: Literal["az", "el"]) -> None:
        ...

    @abstractmethod
    def get_step(self, axis: Literal["az", "el"]) -> int:
        ...

    @abstractmethod
    def get_speed(self, axis: Literal["az", "el"]) -> float:
        ...

    @abstractmethod
    def finalize(self) -> None:
        ...



from types import ModuleType
from typing import Any, Dict, List

from .. import config, get_logger, utils
from ..exceptions import ConfigurationError

logger = get_logger(__name__)


def parse_device_configuration(modules: List[ModuleType]) -> Dict[str, Any]:
    implementations = list_implementations(modules)

    devices = config.dev
    if devices is None:
        return {}

    parsed = {}
    for k, v in devices.__dict__.items():
        if v.lower() in implementations.keys():
            parsed[k] = implementations[v.lower()]
            parsed[utils.toCamelCase(k)] = implementations[v.lower()]
        else:
            raise ConfigurationError(
                f"Driver implementation for device '{v}' ({k}) not found."
            )
    return parsed


def list_implementations(modules: List[ModuleType]) -> Dict[str, Any]:
    def list_implementations_single_module(module: ModuleType) -> Dict[str, Any]:
        impl = {}
        for attrname in dir(module):
            attr = getattr(module, attrname)
            if (
                hasattr(attr, "Model")
                and hasattr(attr, "Manufacturer")
                and callable(attr)
            ):
                impl[attrname.lower()] = attr

        return impl

    implementations = {}
    for module in modules:
        impl = list_implementations_single_module(module)
        implementations.update(impl)
    if len(implementations) != len(set(implementations.keys())):
        raise ValueError("Implemented device model isn't unique.")
    return implementations



__all__ = ["CPZ7415V"]

import queue
import time
from threading import Event, Thread
from typing import Literal, Tuple

from .pulse_controller_base import PulseController
from ...exceptions import ConfigurationError
from ... import config, get_logger, utils


class CPZ7415V(PulseController):
    """Pulse controller, which can handle up to 4 motors (axes).

    - rsw_id
        Board identifier, set by rotary switch with label "RSW1" mounted on the side of
        the board. The board is shipped with default RSW1 setting of 0. This ID would be
        non-zero, when multiple PCI board of same model are mounted on a single FA
        (Factory Automation) controller. Accepted values are ``[0, 1, ..., 16]`` if
        given as ``int``, ``["0", "1", ..., "9", "A", ..., "F"]`` if ``str``.
    - other parameters
        See <http://www.interface.co.jp/download/tutorial/tut0053_14.pdf>.

    """

    Manufacturer = "Interface"
    Model = "CPZ7415V"

    CommandQueueMaxSize = {"warning": 10, "fatal": 30}

    def __init__(self) -> None:
        if config.antenna_cpz7415v is None:
            raise ConfigurationError("Parameters for CPZ-7415V not configured at all.")

        self.logger = get_logger(__name__)

        self.rsw_id = int(config.antenna_cpz7415v_rsw_id)
        self.do_status = config.antenna_cpz7415v_do_conf
        self.use_axes = config.antenna_cpz7415v_useaxes.lower()
        self.axis_mapping = config.antenna_cpz7415v_axis.__dict__
        _config = {
            ax: getattr(config, f"antenna_cpz7415v_{ax}") for ax in self.use_axes
        }

        self.motion = {
            ax: getattr(config, f"antenna_cpz7415v_{ax}_motion").__dict__
            for ax in self.use_axes
        }

        self.current_speed = {ax: 0 for ax in self.use_axes}
        self.current_step = {ax: 0 for ax in self.use_axes}
        self.current_moving = {ax: 0 for ax in self.use_axes}
        self.last_direction = {ax: 0 for ax in self.use_axes}

        self.mode = [conf.mode for conf in _config.values()]  # List type, for utility.
        self.move_mode = {ax: conf.mode for ax, conf in _config.items()}
        self.pulse_conf = {ax: conf.pulse_conf for ax, conf in _config.items()}
        self.default_speed = {ax: conf.motion_speed for ax, conf in _config.items()}
        self.low_speed = {ax: conf.motion_low_speed for ax, conf in _config.items()}

        self.task_queue = queue.Queue()

        self.io = self._initialize_io()
        self.thread = self._start_background_process()

    @utils.skip_on_simulator
    def _initialize_io(self):
        import pyinterface

        io = pyinterface.open(7415, self.rsw_id)
        for ax in self.use_axes:
            io.set_pulse_out(ax, "method", [self.pulse_conf[ax]])
        io.set_motion(self.use_axes, self.mode, self.motion)
        io.output_do([1, 1, 0, 0])
        return io

    def _start_background_process(self) -> Thread:
        self._stop_event = Event()
        thread = Thread(target=self._consume_task_queue, daemon=True)
        thread.start()
        return thread

    def _stop_background_process(self) -> None:
        event = getattr(self, "_stop_event", None)
        if event is not None:
            event.set()
            self._stop_event = None

    @utils.skip_on_simulator
    def _consume_task_queue(self) -> None:
        while not self._stop_event.is_set():
            speed = self.io.read_speed(self.use_axes)
            step = self.io.read_counter(self.use_axes, cnt_mode="counter")
            moving = [int(s[0]) for s in self.io.driver.get_main_status(self.use_axes)]

            self.current_speed = {ax: v for ax, v in zip(self.use_axes, speed)}
            self.current_step = {ax: v for ax, v in zip(self.use_axes, step)}
            self.current_moving = {ax: v for ax, v in zip(self.use_axes, moving)}

            if not self.task_queue.empty():
                if self._check_realtimeness(self.task_queue):
                    task = self.task_queue.get()
                    task["func"](task["args"], task["axis"])
                else:
                    # Commands in task_queue are too outdated, so it's dangerous to
                    # execute them.
                    _ = [self._stop(None, ax) for ax in self.use_axes]
            time.sleep(1e-5)

    def _check_realtimeness(self, queue_: queue.Queue) -> bool:
        qsize = queue_.qsize()
        if qsize < self.CommandQueueMaxSize["warning"]:
            return True
        elif qsize < self.CommandQueueMaxSize["fatal"]:
            self.logger.warning(
                f"{qsize} commands remain unprocessed. "
                "Consider reducing command frequency `antenna_command_frequency`."
            )
            return True
        else:
            self.logger.fatal(f"{qsize} commands remain unprocessed. Emergency stop...")
            return False

    def _convert_ax(self, axis: Literal["az", "el"]) -> str:
        if axis in self.use_axes:
            return axis
        return self.axis_mapping[axis.lower()]

    def set_step(self, step: int, axis: str) -> None:
        ax = self._convert_ax(axis)
        if self.move_mode[ax] == "ptp":
            if self.current_moving[ax] != 0:
                self.task_queue.put(
                    {"func": self._change_step, "args": step, "axis": ax}
                )
            else:
                speed_step = [self.motion[ax]["speed"], step]
                self.task_queue.put(
                    {"func": self._start, "args": speed_step, "axis": ax}
                )

    def set_speed(self, speed: float, axis: Literal["az", "el"]) -> None:
        ax = self._convert_ax(axis)
        speed *= getattr(config.antenna_speed_to_pulse_factor, axis).to_value("s/deg")

        if abs(speed) < self.low_speed[ax]:
            self.task_queue.put({"func": self._stop, "args": None, "axis": ax})
            self.last_direction[ax] = 0
            return
        if self.move_mode[ax] == "jog":
            if (self.last_direction[ax] * speed > 0) and (self.current_moving[ax] != 0):
                self.task_queue.put(
                    {"func": self._change_speed, "args": abs(speed), "axis": ax}
                )
            else:
                direction = 1 if speed > 0 else -1
                speed_step = [abs(speed), direction]
                self.last_direction[ax] = direction
                self.task_queue.put(
                    {"func": self._start, "args": speed_step, "axis": ax}
                )
                time.sleep(0.01)
        return

    def set_do_status(self, status: Tuple[int, int, int, int], do_num: int) -> None:
        self.do_status[do_num - 1] = status
        self.task_queue.put(
            {"func": self._output_do, "args": self.do_status, "axis": None}
        )

    @utils.skip_on_simulator
    def _stop(self, _, ax: str) -> None:
        self.io.stop_motion(axis=ax, stop_mode="immediate_stop")
        while int(self.io.driver.get_main_status(ax)[0][0]) != 0:
            time.sleep(1e-5)

    @utils.skip_on_simulator
    def _change_step(self, step: int, ax: str) -> None:
        self.io.change_step(axis=ax, step=[step])

    @utils.skip_on_simulator
    def _change_speed(self, speed: float, ax: str) -> None:
        self.io.change_speed(axis=ax, mode="accdec_change", speed=[speed])

    @utils.skip_on_simulator
    def _start(self, args, ax) -> None:
        self._stop(None, ax=ax)
        self.motion[ax]["speed"] = args[0]
        self.motion[ax]["step"] = int(args[1])

        axis_mode = [self.move_mode[ax]]
        self.io.set_motion(axis=ax, mode=axis_mode, motion=self.motion)
        self.io.start_motion(axis=ax, start_mode="const", move_mode=self.move_mode[ax])

    @utils.skip_on_simulator
    def _output_do(self, args, _) -> None:
        self.io.output_do(args)

    def get_speed(self, axis: Literal["az", "el"]) -> float:
        ax = self._convert_ax(axis)
        return self.current_speed[ax]

    def get_step(self, axis: Literal["az", "el"]) -> int:
        ax = self._convert_ax(axis)
        return int(self.current_step[ax])

    def finalize(self) -> None:
        self._output_do([0, 0, 0, 0], None)



__all__ = ["PathFinder"]

import math
import time
from typing import List, Tuple, TypeVar, Union

import astropy.constants as const
import astropy.units as u
from astropy.coordinates import (
    BaseCoordinateFrame,
    EarthLocation,
)
import numpy as np

from .convert import CoordCalculator
from .. import config, get_logger, utils
from ..parameters.pointing_error import PointingError
from ..typing import Number, PathLike


T = TypeVar("T", Number, u.Quantity)  # lon, lat の型


class PathFinder:
    """望遠鏡の軌道を計算する

    Parameters
    ----------
    location
        Location of observatory.
    pointing_param_path
        Path to pointing parameter file.
    pressure
        Atmospheric pressure at the observation environment.
    temperature
        Temperature at the observation environment.
    relative_humidity
        Relative humidity at the observation environment.
    obswl
        Observing wavelength.

    Attributes
    ----------
    location: EarthLocation
        Location of observatory.
    pressure: Quantity
        Atmospheric pressure, to compute diffraction correction.
    temperature: Quantity
        Temperature, to compute diffraction correction.
    relative_humidity: Quantity or float
        Relative humidity, to compute diffraction correction.
    obswl: Quantity
        Observing wavelength, to compute diffraction correction.
    obsfreq: Quantity
        Observing frequency of EM-wave, to compute diffraction correction.

    Examples
    --------
    >>> location = EarthLocation("138.472153deg", "35.940874deg", "1386m")
    >>> path = "path/to/pointing_param.toml"
    >>> pressure = 850 * u.hPa
    >>> temperature = 290 * u.K
    >>> humid = 0.5
    >>> obsfreq = 230 * u.GHz
    >>> finder = neclib.coordinates.PathFinder(
    ...     location, path, pressure=pressure, temperature=temperature,
    ...     relative_humidity=humid, obsfreq=obsfreq)

    """

    def __init__(
        self,
        location: EarthLocation,
        pointing_param_path: PathLike,
        *,  # 以降の引数はキーワード引数（引数名=値）として受け取ることを強制する
        pressure: u.Quantity = None,
        temperature: u.Quantity = None,
        relative_humidity: Union[Number, u.Quantity] = None,
        obswl: u.Quantity = None,
        obsfreq: u.Quantity = None,
    ) -> None:
        self.logger = get_logger(self.__class__.__name__)

        if (obswl is not None) and (obsfreq is not None):
            if obswl != const.c / obsfreq:
                raise ValueError("Specify ``obswl`` or ``obs_freq``, not both.")

        self.location = location
        self.pointing_param_path = pointing_param_path
        self.pressure = pressure
        self.temperature = temperature
        self.relative_humidity = relative_humidity
        self.obswl = obswl
        if temperature is not None:
            self.temperature = temperature.to("deg_C", equivalencies=u.temperature())
        if obsfreq is not None:
            self.obswl = const.c / obsfreq

        self.pointing_error_corrector = PointingError.from_file(pointing_param_path)

        diffraction_params = ["pressure", "temperature", "relative_humidity", "obswl"]
        not_set = list(filter(lambda x: getattr(self, x) is None, diffraction_params))
        if len(not_set) > 0:
            self.logger.warning(
                f"{not_set} are not set. Diffraction correction is not available."
            )

    def linear(
        self,
        start: Tuple[T, T],
        end: Tuple[T, T],
        frame: Union[str, BaseCoordinateFrame],
        speed: Union[float, int, u.Quantity],
        *,
        unit: Union[str, u.Unit] = None,
        # obstime: Union[Number, Time] = None,
    ) -> Tuple[u.Quantity, u.Quantity, List[float]]:  # (Az 配列, El 配列, t 配列)
        """望遠鏡の直線軌道を計算する

        Parameters
        ----------
        start
            Longitude and latitude of start point.
        end
            Longitude and latitude of end point.
        frame
            Coordinate frame, in which longitude and latitude are given.
        speed
            Telescope drive speed.
        unit
            Angular unit in which longitude and latitude are given. If they are given as
            ``Quantity``, this parameter will be ignored.

        Examples
        --------
        >>> finder.linear(
            start=(0, 0), end=(0.05, 0), frame="altaz", speed=0.5, unit=u.deg
        )
        [<Quantity [-1.47920569, -1.46920569, -1.45920569, -1.44920569, -1.43920569,
           -1.42920569] deg>, <Quantity [-1.88176239, -1.88176188, -1.88176136,
           -1.88176084, -1.88176032, -1.8817598 ] deg>, array([1.66685637e+09,
           1.66685637e+09, 1.66685637e+09, 1.66685637e+09,
           1.66685637e+09, 1.66685637e+09])]

        """
        obstime = time.time()
        try:
            start_time = float(obstime)
        except TypeError:
            start_time = (obstime.to("second")).value
        try:
            speed = float(speed)
        except TypeError:
            pass
        else:
            speed *= u.deg / u.second
        frequency = config.antenna_command_frequency
        offset = config.antenna_command_offset_sec
        start_lon, start_lat = utils.get_quantity(start[0], start[1], unit=unit)
        end_lon, end_lat = utils.get_quantity(end[0], end[1], unit=unit)
        required_time = max(abs(end_lon - start_lon), abs(end_lat - start_lat)) / speed
        required_time = (required_time.to("second")).value
        command_num = math.ceil(required_time * frequency)
        required_time = command_num / frequency
        command_num += 1  # 始点の分を追加
        lon = [
            start_lon
            + (i * speed / (frequency / u.second)) * np.sign(end_lon - start_lon)
            for i in range(command_num)
        ]
        lat = [
            start_lat
            + (i * speed / (frequency / u.second)) * np.sign(end_lat - start_lat)
            for i in range(command_num)
        ]
        t = [start_time + offset + i / frequency for i in range(command_num)]
        calculator = CoordCalculator(
            location=self.location,
            pointing_param_path=self.pointing_param_path,
            pressure=self.pressure,
            temperature=self.temperature,
            relative_humidity=self.relative_humidity,
            obswl=self.obswl,
        )
        az, el, _ = calculator.get_altaz(
            lon=lon, lat=lat, frame=frame, unit=unit, obstime=t
        )
        return (az, el, t)



"""Type aliases for simple type hinting."""

__all__ = ["PathLike", "AngleUnit"]

import os
from typing import Literal, Protocol, Union, runtime_checkable

import astropy.units as u
import numpy as np


PathLike = Union[os.PathLike, str]
"""Alias of ``os.PathLike``, with ``str`` type combined."""

AngleUnit = Literal["deg", "arcmin", "arcsec", "rad"]
"""Literal expression of supported angular units in this package."""

Number = Union[int, float, np.ndarray]
"""Number types."""

Boolean = Union[bool, np.bool_]
"""Boolean types."""

Unit = Union[str, u.Unit]
"""Unit of physical quantity."""

QuantityValue = Union[Number, u.Quantity]
"""Physical quantity or primitive type with unit separately specified."""


@runtime_checkable
class TextLike(Protocol):
    def upper(self):
        ...

    def lower(self):
        ...

    def find(self):
        ...

    def replace(self):
        ...


@runtime_checkable
class SupportsComparison(Protocol):
    def __eq__(self, other):
        ...

    def __ne__(self, other):
        ...

    def __lt__(self, other):
        ...

    def __le__(self, other):
        ...

    def __gt__(self, other):
        ...

    def __ge__(self, other):
        ...
